/*  GRES, a stripped down version of GNU SED, which implements the 
 *  substitute command only.
 *
 *  This file was modified from sed.c, as included in the GNU SED
 *  version 2.05 source distribution, on July 29, 1998.
 *
 *  Please read the file CHANGES which accompanies this file for details.
 *
 *  Copyright (C) 1989, 1990, 1991 Free Software Foundation, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#ifdef __STDC__
#    define VOID void
#else /* ifdef __STDC__ */
#    define VOID char
#endif /* ifdef __STDC__ */


#define _GNU_SOURCE
#include <ctype.h>
#ifndef isblank
#    define isblank(c) ((c) == ' ' || (c) == '\t')
#endif /* ifndef isblank */
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#if HAVE_UNISTD_H
#    include <unistd.h>
#    ifndef STDIN_FILENO
#        define STDIN_FILENO 0
#    endif /* ifndef STDIN_FILENO */
#endif /* if HAVE_UNISTD_H */
#include "regex.h"
#include "getopt.h"
#if defined(STDC_HEADERS)
#    include <stdlib.h>
#endif /* defined(STDC_HEADERS) */
#if HAVE_STRING_H || defined(STDC_HEADERS)
#    include <string.h>
#    ifndef bzero
#        define bzero(s, n)    memset((s), 0, (n))
#    endif /* ifndef bzero */
#    if !defined(STDC_HEADERS)
#        include <memory.h>
#    endif /* !defined(STDC_HEADERS) */
#else /* HAVE_STRING_H || defined(STDC_HEADERS) */
#    include <strings.h>
#endif /* if HAVE_STRING_H || defined(STDC_HEADERS) */

#ifndef HAVE_BCOPY
#    ifdef HAVE_MEMCPY
#        define bcopy(FROM,TO,LEN)  memcpy(TO,FROM,LEN)
#    else /* HAVE_MEMCPY */
void
bcopy(char *from, char *to, int len)
{
    if (from < to) {
        from += len - 1;
        to += len - 1;
        while (len--) {
        *to-- = *from--;
        }
    } else {
        while (len--) {
            *to++ = *from++;
        }
    }
}

#    endif /* HAVE_MEMCPY */
#endif /* HAVE_BCOPY */

char *version_string = "GRES version 1.4.5";

#define S_GLOBAL_BIT    01

/* Gres operates a line at a time. */
struct line
{
    char *text;                 /* Pointer to line allocated by malloc. */
    int length;                 /* Length of text. */
    int alloc;                  /* Allocated space for text. */
};


#if defined(__STDC__)
#    define P_(s) s
#else
#    define P_(s) ()
#endif

void panic P_((char *str,...));
char *__fp_name P_((FILE * fp));
FILE *ck_fopen P_((char *name, char *mode));
void ck_fwrite P_((char *ptr, int size, int nmemb, FILE * stream));
void ck_fclose P_((FILE * stream));
void *ck_malloc P_((int size));
void *ck_realloc P_((void * ptr, int size));
char *ck_strdup P_((char *str));
void *init_buffer P_((void));
void flush_buffer P_((void * bb));
int size_buffer P_((void * b));
void add_buffer P_((void * bb, char *p, int n));
void add1_buffer P_((void * bb, int ch));
char *get_buffer P_((void * bb));


void compile_program P_((char *, char *, int));
void bad_prog P_((char *why));
int inchar P_((void));
void savchar P_((int ch));
void buffer_regex P_((void));
void compile_regex P_((void));
void read_file P_((char *name, char *, int, int, int));
int execute_program P_((char *, int, int, int));
int read_pattern_space P_((void));
void append_pattern_space P_((void));
void str_append P_((struct line * to, char *string, int length));
void usage P_((int));
size_t block_read P_((void));

extern char *myname;

char * last_regex_string = 0;

/* If set, don't write out the line unless a replacement is made */
int no_default_output = 0;

/* Current input line # */
/* int input_line_number = 0; */

/* Points to current position in memory buffer */
size_t position = 0;

/* Points to the end of the file in the memory buffer */
size_t end_of_data = 0;

/* Size of data blocks to read */
const size_t block_size = BUFSIZ;

/* Buffer to store block reads */
unsigned char *static_buffer;

/* The size of the buffer we allocate when searching a line */
int alloc_len = 50; /* MIN_ALLOCATE is #define'd in utils.c to 50 */

/* The size of the memory area to malloc during fastmap in compile_regex */
int fastmap_size = 256;

/* The 'current' input line. */
struct line line;

/* A 'line' to append to the current line when it comes time to write it out */
struct line append;

/* When we're reading a script command from a string, 'prog_start' and
 * 'prog_end' point to the beginning and end of the string.  This
 * would allow us to compile script strings that contain nulls, except
 * that script strings are only read from the command line, which is
 * null-terminated */
unsigned 
char *prog_start;
unsigned 
char *prog_end;

/* When we're reading a script command from a string, 'prog_cur' points
 * to the current character in the string */
unsigned 
char *prog_cur;

/* This is the file pointer that we're currently reading data from.  It may
 * be stdin. */
int input_file;

/* If this variable is non-zero at exit, one or more of the input
 * files couldn't be opened. */
int bad_input = 0;

/* 'an empty regular expression is equivalent to the last regular
 * expression read' so we have to keep track of the last regex used.
 * Here's where we store a pointer to it (it is only malloc()'d once) */
struct 
re_pattern_buffer *last_regex;

/* Various error messages we may want to print */
static 
char BAD_EOF[] = "Unexpected End-of-file";
static 
char NO_REGEX[] = "No previous regular expression";

static struct re_registers regs =
{0, 0, 0};

static 
struct option longopts[] =
{
    {"version", 0, NULL, 'V'},
    {"global", 0, NULL, 'g'},
    {"pattern", 1, NULL, 'p'},
    {"replace", 1, NULL, 'r'},
    {"no-default", 0, NULL, 'n'},
    {"help", 0, NULL, 'h'},
    {NULL, 0, NULL, 0}
};

int
main(int argc, char **argv)
{
    int opt;
    int compiled = 0;
    int required_arguments = 2; /* Unless -p and/or -r is given */
    char *pattern = NULL;
    char *replace = NULL;
    int replacement_len = 0;
    int pattern_len = 0;
    int global = 0;

    /* see regex.h */
    re_set_syntax(RE_SYNTAX_POSIX_BASIC);

    myname = argv[0];
    while ((opt = getopt_long(argc, argv, "hngp:r:V", longopts, (int *) 0))
            != EOF) {
        switch (opt) {
            case 'g':
                global = 1;
                break;
            case 'p':
                if (pattern == NULL) {
                    required_arguments--;
                } else {
                    free(pattern);
                }
                pattern = ck_malloc(strlen(optarg) + 1);
                strcpy(pattern, optarg);
                break;
            case 'r':
                if (replace == NULL) {
                    required_arguments--;
                } else {
                    free(replace);
                }
                replace = ck_malloc(strlen(optarg) + 1);
                strcpy(replace, optarg);
                break;
            case 'V':
                fprintf(stderr, "%s\n", version_string);
                exit(0);
                break;
            case 'n':
                no_default_output = 1;
                break;
            case 'h':
                usage(0);
                break;
            default:
                usage(4);
                break;
        }
    }
    /* If there are too few arguments to gres, output error message and
     * exit. */
    if ((argc - optind) < required_arguments) {
        usage(4);
    }
    /* Allocate memory for the commands, plus 1 (null) plus 1 (separator 
     * between the pattern and the replacement). */
    if (pattern == NULL) {
        pattern = ck_malloc(strlen (argv[optind]) + 1);
        strcpy(pattern, argv[optind++]);
    }
    if (replace == NULL) { 
        replace = ck_malloc(strlen(argv[optind]) + 1);
        strcpy(replace, argv[optind++]);
    }
    replacement_len = strlen(replace);
    pattern_len = strlen(pattern);

    compile_program(pattern, replace, replacement_len);

    line.length = 0;
    line.alloc = alloc_len;
    line.text = ck_malloc(alloc_len);

    append.length = 0;
    append.alloc = alloc_len;
    append.text = ck_malloc(alloc_len);

/* read_file invokes execute_program */
    if (argc <= optind) {
        read_file("-", replace, replacement_len, pattern_len, global);
    } else {
        while (optind < argc) {
            read_file(argv[optind], replace, replacement_len, pattern_len, 
                      global);
            optind++;
        }
    }
    /* Return memory from temporary variables and assembled command. */
    free(pattern);
    free(replace);
    if (bad_input) {
        exit (2);
    }
    exit (0);
}

void
compile_program(char *patt, char *repl, int repl_len)
{
    int ch = 0;

    prog_start = prog_cur = patt;
    prog_end = patt + strlen(patt);

    for (;;) {
        ch = inchar(); 
        if (ch == EOF) {
            break;
        }
        savchar(ch);

        buffer_regex();
        compile_regex();

        prog_start = repl;
        prog_end = repl + repl_len;
        prog_cur = 0;
    }
    return;
}

/* Complain about an error and exit. */
void
bad_prog(char *why)
{
    fprintf(stderr, "%s: %s\n", myname, why);
    exit(1);
}

/* Read the next character from the program.    Return EOF if there isn't
 * anything to read.    */
int
inchar()
{
    int ch;

    if (!prog_cur) {
        return EOF;
    } else if (prog_cur == prog_end) {
        ch = EOF;
        prog_cur = 0;
    } else {
        ch = *prog_cur++;
    }
    return ch;
}

/* unget 'ch' so the next call to inchar will return it.    'ch' must not be
 * EOF or anything nasty like that. */
void
savchar(int ch)
{
    if (ch == EOF) {
        return;
    }
    *--prog_cur = ch;
}

void
buffer_regex()
{
    VOID *b;
    int ch;

    b = init_buffer();
    while ((ch = inchar()) != EOF) {
        if (ch == '^') {
            if (size_buffer(b) == 0) {
                add1_buffer(b, '\\');
                add1_buffer(b, '`');
            } else {
                add1_buffer(b, ch);
            }
            continue;
        } else if (ch == '[') {
            add1_buffer(b, ch);
            continue;
        } else if (ch == ']') {
            add1_buffer(b, ch);
            {
                char * regexp = get_buffer(b);
                int pos = size_buffer(b) - 1;
                continue;
            }
        } else if (ch != '\\') {
            add1_buffer(b, ch);
            continue;
        }
        ch = inchar();
        switch (ch) {
            case 'n':
                add1_buffer(b, '\n');
                break;
            case EOF:
                break;
            default:
                add1_buffer(b, '\\');
                add1_buffer(b, ch);
                break;
        }
    }
    if (size_buffer(b)) {
        if (last_regex_string) {
            last_regex_string = (char *)ck_realloc(last_regex_string,
                                                   size_buffer(b) + 1);
        } else {
            last_regex_string = (char *)ck_malloc(size_buffer(b) + 1);
        }
        bcopy(get_buffer(b), last_regex_string, size_buffer(b));
        last_regex_string[size_buffer(b)] = 0;
    } else if (!last_regex) {
        bad_prog(NO_REGEX);
    }
    flush_buffer(b);
}

void
compile_regex()
{
    const char * error;
    last_regex = ((struct re_pattern_buffer *) 
            ck_malloc(sizeof(struct re_pattern_buffer)));
    bzero(last_regex, sizeof(*last_regex));
    last_regex->fastmap = ck_malloc(fastmap_size);
    error = re_compile_pattern(last_regex_string,
                               strlen(last_regex_string), last_regex);
    if (error) {
        bad_prog((char *)error);
    }
}

/* Read a file and apply the compiled script to it. */
void
read_file(char *name, char *replacement, int replace_len, int patt_len, 
          int glob)
{
    int replaced = 0;

    if (*name == '-' && name[1] == '\0') {
        input_file = STDIN_FILENO;
    } else {
        input_file = open(name, O_RDONLY);
        if (input_file == -1) {
            extern int errno;
#if !defined(FREEBSD) && ((__GNU_LIBRARY__ - 0 < 6) || defined(USE_SFIO)) 
            extern char *sys_errlist[];
#endif
            extern int sys_nerr;
            const char *ptr;
            if ((errno >= 0) && (errno < sys_nerr)) {
                ptr = sys_errlist[errno];
            } else {
                ptr = "Unknown error code";
            }
            bad_input++;
            fprintf(stderr, "%s: can't read %s: %s\n", myname, name, ptr);
            return;
        }
    }

/* open a buffer into which the opened file will be read */
    static_buffer = ck_malloc(block_size + 1);
    while (read_pattern_space()) {
        replaced = execute_program(replacement, replace_len, patt_len, glob);
        if ((!no_default_output) || (replaced)) {
            ck_fwrite(line.text, 1, line.length, stdout);
        }
        if (append.length) {
            ck_fwrite(append.text, 1, append.length, stdout);
            append.length = 0;
        }
    } 
    close(input_file);
    free(static_buffer);
}

/* Execute the program on the current input line. */
int
execute_program(char *rep, int rep_len, int pat_len, int global_bit)
{
    int n;
    static int end_cycle;

    int start;
    int remain;
    int offset;

    static struct line tmp;
    struct line t;
    char *rep_end, *rep_next, *rep_cur;

    int count;
    int replaced_num = 0;

    count = 0;
    start = 0;

    end_cycle = 0;

    if (!tmp.alloc) {
        tmp.alloc = alloc_len;
        tmp.text = ck_malloc(alloc_len);
    }
    remain = line.length;
    tmp.length = 0;
    rep_end = rep + rep_len;

    while ((offset = re_search(last_regex, line.text, line.length, start,
                               remain, &regs)) >= 0) {
        count++;
        if (offset - start) {
            str_append(&tmp, line.text + start, offset - start);
        }
        for (rep_next = rep_cur = rep; rep_next < rep_end; rep_next++) {
            if (*rep_next == '&') {
                if (rep_next - rep_cur) {
                    str_append(&tmp, rep_cur, rep_next - rep_cur);
                }
                str_append(&tmp, line.text + regs.start[0], 
                           regs.end[0] - regs.start[0]);
                rep_cur = rep_next + 1;
            } else if (*rep_next == '\\') {
                if (rep_next - rep_cur) {
                    str_append(&tmp, rep_cur, rep_next - rep_cur);
                }
                rep_next++;
                if (rep_next != rep_end) {
                    if (*rep_next >= '0' && *rep_next <= '9') {
                        n = *rep_next - '0';
                        str_append(&tmp, line.text + regs.start[n], 
                                   regs.end[n] - regs.start[n]);
                    } else {
                        str_append(&tmp, rep_next, 1);
                    }
                }
                rep_cur = rep_next + 1;
            }
        }
        if (rep_next - rep_cur) {
            str_append(&tmp, rep_cur, rep_next - rep_cur);
        }
        if (offset == regs.end[0]) {
            str_append(&tmp, line.text + offset, 1);
            ++regs.end[0];
        }
        start = regs.end[0];

        remain = line.length - start;
        if (remain < 0) {
            break;
        }
        if (!(global_bit & S_GLOBAL_BIT)) {
            break;
        }
    }
    if (!count) {
        return(0);
    }
    str_append(&tmp, line.text + start, remain);
    t.text = line.text;
    t.length = line.length;
    t.alloc = line.alloc;
    line.text = tmp.text;
    line.length = tmp.length;
    line.alloc = tmp.alloc;
    tmp.text = t.text;
    tmp.length = t.length;
    tmp.alloc = t.alloc;
    replaced_num++;
    return (replaced_num);
}

/* Read in the next line of input, and store it in the pattern space.    Return 
 * non-zero if this is the last line of input. */

int
read_pattern_space()
{
    int n;
    char *p;
    char ch = 0;

    p = line.text;
    n = line.alloc;


    if (position == 0) {
        end_of_data = block_read();
    }
    if (end_of_data == -1) {
        fprintf(stderr, "Error reading file.\n");
        return -1;
    }

/*    input_line_number++; */
    for (;;) {
        while (position < end_of_data) {
            if (n == 0) {
                line.text = ck_realloc(line.text, line.alloc * 2);
                p = line.text + line.alloc;
                n = line.alloc;
                line.alloc *= 2;
            }
            ch = static_buffer[position];
            position++;
            *p++ = ch;
            --n;
            if (ch == '\n') {
                line.length = line.alloc - n;
                break;
            }
        }
        if (ch == '\n') {
            break;
        }
        if (end_of_data == 0) {
            if (n == line.alloc) {
                return 0;
            }
            line.length = line.alloc - n;
            return 1;
        } else {
            end_of_data = block_read();
            if (end_of_data == -1) {
                fprintf(stderr, "Error reading file.\n");
                return -1;
            }
            position = 0;
            ch = 0;
        }
    }
    return 1;
}

/* Block read file stream 
 * block_read takes a memory buffer which was previously allocated dynamically
 * with open_memstream and a size_t count which specifies how big the buffer
 * is (as well as what size block to read). */
size_t
block_read()
{
    size_t pos = 0;

/* This should read block_size records of size 1 byte from input_file and 
 * store the data in the memory buffer.    Pos holds the number of records 
 * read in the memory buffer. This routine will use a memory buffer of 
 * block_size static size and read in a variable amount of data. */
    pos = read(input_file, static_buffer, block_size);
    return(pos);
}
    
/* Append 'length' bytes from 'string' to the line 'to'.    This routine *will*
 * append bytes with nulls in them, without failing. */
void
str_append(struct line *to, char *string, int length)
{
    if (length > to->alloc - to->length) {
        to->alloc += length;
        to->text = ck_realloc(to->text, to->alloc);
    }
    bcopy(string, to->text + to->length, length);
    to->length += length;
}

void
usage(int status)
{
    fprintf(status ? stderr : stdout, 
        "Usage: %s [-V, --version] [-g, --global] [-n, --no-default]\n\t [-p, --pattern] pattern [-r, --replace] replacement [file...]\n", 
        myname);
    exit(status);
}
